package __

import static com.kms.katalon.core.checkpoint.CheckpointFactory.findCheckpoint
import static com.kms.katalon.core.testcase.TestCaseFactory.findTestCase
import static com.kms.katalon.core.testdata.TestDataFactory.findTestData
import static com.kms.katalon.core.testobject.ObjectRepository.findTestObject
import static com.kms.katalon.core.testobject.ObjectRepository.findWindowsObject
import static groovy.io.FileType.FILES

import com.google.api.client.util.Charsets
import com.helger.commons.io.stream.StringInputStream
import com.kms.katalon.core.annotation.Keyword
import com.kms.katalon.core.checkpoint.Checkpoint
import com.kms.katalon.core.cucumber.keyword.CucumberBuiltinKeywords as CucumberKW
import com.kms.katalon.core.mobile.keyword.MobileBuiltInKeywords as Mobile
import com.kms.katalon.core.model.FailureHandling
import com.kms.katalon.core.testcase.TestCase
import com.kms.katalon.core.testdata.TestData
import com.kms.katalon.core.testobject.TestObject
import com.kms.katalon.core.webservice.keyword.WSBuiltInKeywords as WS
import com.kms.katalon.core.webui.keyword.WebUiBuiltInKeywords as WebUI
import com.kms.katalon.core.windows.keyword.WindowsBuiltinKeywords as Windows
import com.kms.katalon.core.configuration.RunConfiguration
import com.kms.katalon.core.util.KeywordUtil


import internal.GlobalVariable

import org.apache.commons.lang.NotImplementedException
import org.apache.poi.openxml4j.exceptions.InvalidOperationException
import groovy.json.JsonSlurper
import groovy.json.internal.LazyMap
import groovy.json.JsonBuilder
import groovy.lang.GroovyClassLoader
import groovy.lang.Tuple2
import groovy.util.slurpersupport.GPathResult
import groovy.xml.XmlUtil
import java.time.LocalDate as LocalDate
import java.time.LocalTime as LocalTime
import java.time.LocalDateTime as LocalDateTime
import java.time.Duration as Duration
import java.time.ZonedDateTime as ZonedDateTime
import java.time.Period as Period
import java.lang.Class
import java.nio.charset.Charset
import org.apache.commons.lang.StringEscapeUtils

public class Data {

	public static HashMap parse(String testCaseId, HashMap testCaseVariables, ITestDataParser parser = new TestDataParser("__")) {
		def result = new HashMap()

		ArrayList<TestDataInfo> testDataInfoList = getTestDataInfoList(testCaseId, testCaseVariables)

		for(testDataInfo in testDataInfoList) {
			result.put(testDataInfo.Name, parser.parseTestData(testDataInfo.Type, testDataInfo.Value))
		}
		
		return result
	}

	private static ArrayList<TestDataInfo> getTestDataInfoList(String testCaseId, HashMap testCaseVariables) {
		GPathResult xml = new XmlSlurper().parse(new File(RunConfiguration.getProjectDir() + "/" + testCaseId + ".tc"))

		def result = new ArrayList<TestDataInfo>()

		xml.variable.each { variable ->
			result.add(new TestDataInfo(variable.name.toString(), variable.description.toString(), testCaseVariables[variable.name.toString()].toString()))
		}

		return result
	}
}

public interface ITestDataParser {
	public Object parseTestData(String type, String value)
}

public class TestDataParser implements ITestDataParser {
	public String _packageNameOfCustomTestDataClasses

	public TestDataParser(String packageNameOfCustomTestDataClasses) {
		_packageNameOfCustomTestDataClasses = packageNameOfCustomTestDataClasses
	}

	public Object parseTestData(String type, String value) {
		
		if(value == '' || value == null || value == 'null')
			return null

		type = type.trim()

		if(type == 'String' || type == '') {
			return value
		}
		else {
			value = JsonHelper.removeLeadingAndTrailingQuotationMarks(value)
			if(type == 'Resource') {
				return RunConfiguration.getProjectDir() + "/" + "Resources" + "/" + value
			}
			else if(type == 'Bool') {
				return value.toLowerCase() == 'true'
			}
			else if(type == 'Char') {
				return value.toCharacter()
			}
			else if(type == 'Byte') {
				return Byte.parseByte(value)
			}
			else if(type == 'Short') {
				return Short.parseShort(value)
			}
			else if(type == 'Int') {
				return Integer.parseInt(value)
			}
			else if(type == 'Long') {
				return Long.parseLong(value)
			}
			else if(type == 'Float') {
				return Float.parseFloat(value)
			}
			else if(type == 'Double') {
				return Double.parseDouble(value)
			}
			else if(type == 'LocalDate') {
				return LocalDate.parse(value)
			}
			else if(type == 'LocalTime') {
				return LocalTime.parse(value)
			}
			else if(type == 'LocalDateTime') {
				return LocalDateTime.parse(value)
			}
			else if(type == 'ZonedDateTime') {
				return ZonedDateTime.parse(value)
			}
			else if(type == 'Period') {
				return Period.parse(value)
			}
			else if(type == 'Duration') {
				return Duration.parse(value)
			}
			else if(type.startsWith('List<')) {
				return parseList(type, value)
			}
			else if(type.startsWith('Map<')) {
				return parseMap(type, value)
			}
			else if(isCustomTestData(type)) {
				return parseCustomTestData(type, value)
			}
			else {
				 throw new InvalidOperationException("Unsupported data type ${type} during parsing")
			}
		}
	}

	private Object parseList(String type, String value) {
		def typeOfElement = type.substring(5, type.length() - 1)

		def arrayList = JsonHelper.parse(value)

		if(!(arrayList instanceof ArrayList)) {
			throw new InvalidOperationException("Can't parse List<${typeOfElement}>, because it is in wrong format")
		}

		def result = new ArrayList()

		for(o in arrayList) {
			result.add(parseTestData(typeOfElement, JsonHelper.serialize(o)))
		}

		return result
	}

	private Object parseMap(String type, String value) {
		def typesOfMap = getTypesOfMap(type)

		def typeOfKey = typesOfMap[0]
		def typeOfValue = typesOfMap[1]

		def map = JsonHelper.parse(value)

		if(!(map instanceof LazyMap)) {
			throw new InvalidOperationException("Can't parse Map<${typeOfKey}, ${typeOfValue}>, because it is in wrong format")
		}

		def result = new HashMap()

		for(pair in map) {
			def parsedKey = parseTestData(typeOfKey, JsonHelper.serialize(pair.key))
			def parsedValue = parseTestData(typeOfValue, JsonHelper.serialize(pair.value))

			result.put(parsedKey, parsedValue)
		}

		return result
	}

	private Tuple2<String, String> getTypesOfMap(String type) {
		def typeWithoutSurroundingMap = type.substring(4, type.length() - 1)

		Integer squareBracketLevel = 0
		Integer signBracketLevel = 0

		Boolean parsingFirstType = true
		StringBuilder firstType = new StringBuilder()
		StringBuilder secondType = new StringBuilder()

		for(char c in typeWithoutSurroundingMap) {
			if(c == ',' && squareBracketLevel == 0 && signBracketLevel == 0) {
				parsingFirstType = false
			}
			else {
				if(c == '<') {
					signBracketLevel++
				}
				else if(c == '>') {
					signBracketLevel--
				}
				else if(c == '[') {
					squareBracketLevel++
				}
				else if(c == ']') {
					squareBracketLevel--
				}

				if(parsingFirstType) {
					firstType.append(c)
				}
				else {
					secondType.append(c)
				}
			}
		}

		return new Tuple2<String, String>(firstType.toString().trim(), secondType.toString().trim())
	}

	private Boolean isCustomTestData(String type) {
		try {
			def foundClass = Class.forName(_packageNameOfCustomTestDataClasses + "." + type)

			def testDataInterface = foundClass as ICustomTestData

			if(testDataInterface != null) {
				return true
			}
			else {
				return false
			}
		}
		catch(Exception ex) {
			return false
		}
	}

	private Object parseCustomTestData(String type, String value) {
		try {
			def instance = new GroovyClassLoader().loadClass(_packageNameOfCustomTestDataClasses + "." + type).newInstance()

			return (instance as ICustomTestData).parseTestData(_packageNameOfCustomTestDataClasses, type, value)
		}
		catch(Exception ex) {
			throw new InvalidOperationException("Can't parse custom data of type ${type}")
		}
	}
}

public interface ICustomTestData {
	public Object parseTestData(String packageName, String type, String value)
}

public abstract class JsonTestData implements ICustomTestData {
	public Object parseTestData(String packageName, String type, String value) {

		def map = JsonHelper.parse(value)

		return new GroovyClassLoader().loadClass(packageName + "." + type).newInstance(map)
	}
}

public class TestDataInfo {
	public String Name
	public String Type
	public String Value

	public TestDataInfo(String name, String type, String value) {
		Name = name
		Type = type
		Value = value
	}
}

public class JsonHelper {
	public static Object parse(String value) {
		InputStream stream = new StringInputStream(removeLeadingAndTrailingQuotationMarks(value), Charsets.UTF_8)
		return new JsonSlurper().parse(stream, 'UTF-8')
	}

	public static String serialize(Object object) {
		String json = removeLeadingAndTrailingQuotationMarks(new JsonBuilder(object).toString())
		return StringEscapeUtils.unescapeJava(json)
	}

	public static String removeLeadingAndTrailingQuotationMarks(String value) {
		return value.replaceAll('^"+', "").replaceAll('"+$', "")
	}
}